#ifndef MSP_GEOMETRY_HALFSPACE_H_
#define MSP_GEOMETRY_HALFSPACE_H_

#include "shape.h"

namespace Msp {
namespace Geometry {

/**
An unbounded shape consisting of the space on one side of a plane.  Mostly
useful when intersected with other shapes.
*/
template<typename T, unsigned D>
class HalfSpace: public Shape<T, D>
{
private:
	LinAl::Vector<T, D> normal;

public:
	HalfSpace();
	HalfSpace(const LinAl::Vector<T, D> &);

	virtual HalfSpace *clone() const;

	const LinAl::Vector<T, D> &get_normal() const { return normal; }

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box(unsigned = 0) const;
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return 1; }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
	virtual Coverage get_coverage(const BoundingBox<T, D> &) const;
};

template<typename T, unsigned D>
inline HalfSpace<T, D>::HalfSpace()
{
	normal[0] = T(1);
}

template<typename T, unsigned D>
inline HalfSpace<T, D>::HalfSpace(const LinAl::Vector<T, D> &n):
	normal(normalize(n))
{ }

template<typename T, unsigned D>
inline HalfSpace<T, D> *HalfSpace<T, D>::clone() const
{
	return new HalfSpace<T, D>(normal);
}

template<typename T, unsigned D>
inline BoundingBox<T, D> HalfSpace<T, D>::get_axis_aligned_bounding_box(unsigned) const
{
	// XXX If the normal is aligned to an axis, should the bounding box reflect that?
	return ~BoundingBox<T, D>();
}

template<typename T, unsigned D>
inline bool HalfSpace<T, D>::contains(const LinAl::Vector<T, D> &point) const
{
	return inner_product(point, normal)<=T(0);
}

template<typename T, unsigned D>
inline unsigned HalfSpace<T, D>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	T d = inner_product(ray.get_start(), normal);
	T c = inner_product(ray.get_direction(), normal);
	if(c==T(0))
		return 0;

	T x = -d/c;
	if(ray.check_limits(x))
	{
		if(points && size>0)
		{
			points[0].position = ray.get_start()+ray.get_direction()*x;
			points[0].normal = normal;
			points[0].distance = x;
			points[0].entry = (c<T(0));
		}

		return 1;
	}

	return 0;
}

template<typename T, unsigned D>
inline Coverage HalfSpace<T, D>::get_coverage(const BoundingBox<T, D> &bbox) const
{
	LinAl::Vector<T, D> low_point;
	LinAl::Vector<T, D> high_point;
	for(unsigned i=0; i<D; ++i)
	{
		low_point[i] = (normal[i]>=T(0) ? bbox.get_minimum_coordinate(i) : bbox.get_maximum_coordinate(i));
		high_point[i] = (normal[i]>=T(0) ? bbox.get_maximum_coordinate(i) : bbox.get_minimum_coordinate(i));
	}

	if(contains(high_point))
		return FULL_COVERAGE;
	else if(contains(low_point))
		return PARTIAL_COVERAGE;
	else
		return NO_COVERAGE;
}

} // namespace Geometry
} // namespace Msp

#endif
