#ifndef MSP_LINAL_MATRIX_H_
#define MSP_LINAL_MATRIX_H_

#include <algorithm>
#include <ostream>
#include "matrixops.h"
#include "vector.h"

namespace Msp {
namespace LinAl {

/**
A general mathematical matrix with M rows and N columns.
*/
template<typename T, unsigned M, unsigned N>
class Matrix
{
public:
	typedef T ElementType;

private:
	T data[M*N];

public:
	Matrix();
	Matrix(const T *);
	template<typename U>
	Matrix(const Matrix<U, M, N> &);

	static Matrix identity();
	static Matrix from_columns(const Vector<T, M> *);
	static Matrix from_rows(const Vector<T, N> *);

	unsigned rows() const { return M; }
	unsigned columns() const { return N; }

	T &element(unsigned i, unsigned j) { return data[i+M*j]; }
	const T &element(unsigned i, unsigned j) const { return data[i+M*j]; }
	T &operator()(unsigned i, unsigned j) { return element(i, j); }
	const T &operator()(unsigned i, unsigned j) const { return element(i, j); }

	Vector<T, M> column(unsigned i) const { return Vector<T, M>(data+M*i); }
	Vector<T, N> row(unsigned i) const { return Vector<T, N>(data+i, M); }

	template<unsigned P, unsigned Q>
	Matrix<T, P, Q> select(const Vector<unsigned, P> &, const Vector<unsigned, Q> &) const;

	template<unsigned P, unsigned Q>
	Matrix<T, P, Q> block(unsigned, unsigned) const;

	Matrix &operator*=(T);
	Matrix &operator*=(const Matrix &);
	Matrix &operator/=(T);
	Matrix &operator+=(const Matrix &);
	Matrix &operator-=(const Matrix &);

	Matrix &invert();

	Matrix &exchange_columns(unsigned, unsigned);
	Matrix &multiply_column(unsigned, T);
	Matrix &add_column(unsigned, unsigned, T);
};

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N>::Matrix()
{
	std::fill(data, data+M*N, T());
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N>::Matrix(const T *d)
{
	std::copy(d, d+M*N, data);
}

template<typename T, unsigned M, unsigned N>
template<typename U>
inline Matrix<T, M, N>::Matrix(const Matrix<U, M, N> &other)
{
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			element(i, j) = other(i, j);
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> Matrix<T, M, N>::identity()
{
	static_assert(M==N, "An identity matrix must be square");
	Matrix<T, M, N> m;
	for(unsigned i=0; i<M; ++i)
		m(i, i) = T(1);
	return m;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> Matrix<T, M, N>::from_columns(const Vector<T, M> *v)
{
	Matrix<T, M, N> m;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			m(i, j) = v[j][i];
	return m;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> Matrix<T, M, N>::from_rows(const Vector<T, N> *v)
{
	Matrix<T, M, N> m;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			m(i, j) = v[i][j];
	return m;
}

template<typename T, unsigned M, unsigned N>
template<unsigned P, unsigned Q>
inline Matrix<T, P, Q> Matrix<T, M, N>::select(const Vector<unsigned, P> &row_indices, const Vector<unsigned, Q> &col_indices) const
{
	Matrix<T, P, Q> r;
	for(unsigned j=0; j<P; ++j)
		for(unsigned i=0; i<Q; ++i)
			r(j, i) = element(row_indices[j], col_indices[i]);
	return r;
}

template<typename T, unsigned M, unsigned N>
template<unsigned P, unsigned Q>
inline Matrix<T, P, Q> Matrix<T, M, N>::block(unsigned y, unsigned x) const
{
	Matrix<T, P, Q> r;
	for(unsigned j=0; j<P; ++j)
		for(unsigned i=0; i<Q; ++i)
			r(j, i) = element(y+j, x+i);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator*=(T s)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] *= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator*=(const Matrix<T, M, N> &m)
{
	static_assert(M==N, "Multiplication-assignment is only possible on square matrices");
	return *this = *this*m;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator*(const Matrix<T, M, N> &m, T s)
{
	Matrix<T, M, N> r(m);
	return r *= s;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator*(T s, const Matrix<T, M, N> &m)
{
	return m*s;
}

template<typename T, unsigned M, unsigned P, unsigned N>
inline Matrix<T, M, N> operator*(const Matrix<T, M, P> &m1, const Matrix<T, P, N> &m2)
{
	Matrix<T, M, N> r;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			for(unsigned k=0; k<P; ++k)
				r(i, j) += m1(i, k)*m2(k, j);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Vector<T, M> operator*(const Matrix<T, M, N> &m, const Vector<T, N> &v)
{
	Vector<T, M> r;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			r[i] += m(i, j)*v[j];
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Vector<T, N> operator*(const Vector<T, M> &v, const Matrix<T, M, N> &m)
{
	Vector<T, N> r;
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			r[j] += v[i]*m(i, j);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator/=(T s)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] /= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator/(const Matrix<T, M, N> &m, T s)
{
	Matrix<T, M, N> r(m);
	return r /= s;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator+=(const Matrix<T, M, N> &m)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] += m.data[i];
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator+(const Matrix<T, M, N> &m1, const Matrix<T, M, N> &m2)
{
	Matrix<T, M, N> r(m1);
	return r += m2;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator-=(const Matrix<T, M, N> &m)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] -= m.data[i];
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator-(const Matrix<T, M, N> &m1, const Matrix<T, M, N> &m2)
{
	Matrix<T, M, N> r(m1);
	return r -= m2;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N>& Matrix<T, M, N>::invert()
{
	static_assert(M==N, "Inversion is only possible on square matrices");
	Matrix<T, M, N> r = identity();
	gauss_jordan(*this, r);
	return *this = r;
}

template<typename T, unsigned S>
inline Matrix<T, S, S> invert(const Matrix<T, S, S> &m)
{
	Matrix<T, S, S> temp = m;
	Matrix<T, S, S> r = Matrix<T, S, S>::identity();
	return gauss_jordan(temp, r);
}

template<typename T, unsigned M, unsigned N>
inline bool operator==(const Matrix<T, M, N> &a, const Matrix<T, M, N> &b)
{
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			if(a(i, j)!=b(i, j))
				return false;
	return true;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::exchange_columns(unsigned i, unsigned j)
{
	using std::swap;
	for(unsigned k=0; k<M; ++k)
		swap(element(k, i), element(k, j));
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::multiply_column(unsigned i, T s)
{
	for(unsigned k=0; k<M; ++k)
		element(k, i) *= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::add_column(unsigned i, unsigned j, T s)
{
	for(unsigned k=0; k<M; ++k)
		element(k, j) += element(k, i)*s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, N, M> transpose(const Matrix<T, M, N> &m)
{
	Matrix<T, N, M> r;
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			r(j, i) = m(i, j);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline std::ostream &operator<<(std::ostream &s, const Matrix<T, M, N> &m)
{
	s << "Matrix" << M << 'x' << N << '(';
	for(unsigned i=0; i<N; ++i)
	{
		if(i)
			s << ", ";
		s << '[';
		for(unsigned j=0; j<M; ++j)
		{
			if(j)
				s << ", ";
			s << m(j, i);
		}
		s << ']';
	}
	s << ')';
	return s;
}

} // namespace LinAl
} // namespace Msp

#endif
