#ifndef MSP_LINAL_VECTOR_H_
#define MSP_LINAL_VECTOR_H_

#include <algorithm>
#include <cmath>
#include <ostream>

namespace Msp {
namespace LinAl {

/**
Base class to provide the components of a vector.  This is used so that
specializations with individual members can be provided in some dimensions.
*/
template<typename T, unsigned N>
class VectorComponents
{
private:
	T data[N];

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return data[i]; }
	const T &operator[](unsigned i) const { return data[i]; }
};

template<typename T>
class VectorComponents<T, 2>
{
public:
	T x, y;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

template<typename T>
class VectorComponents<T, 3>
{
public:
	T x, y, z;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

template<typename T>
class VectorComponents<T, 4>
{
public:
	T x, y, z, w;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

/**
A general mathematical vector.
*/
template<typename T, unsigned N>
class Vector: public VectorComponents<T, N>
{
public:
	typedef T ElementType;

	Vector();
	Vector(const T *);

	/** Constructs a vector from an array of interleaved values.  Intended for
	use by Matrix row accessor. */
	Vector(const T *, unsigned);

	template<typename... Args>
	Vector(T, Args...);

	template<typename U>
	Vector(const Vector<U, N> &);

	unsigned size() const { return N; }

	template<unsigned M>
	Vector<T, M> slice(unsigned) const;

	Vector &operator*=(T);
	Vector &operator/=(T);
	Vector &operator+=(const Vector &);
	Vector &operator-=(const Vector &);

	T norm() const;
	Vector &normalize();
};


template<typename T, unsigned N>
inline Vector<T, N>::Vector()
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = T();
}

template<typename T, unsigned N>
inline Vector<T, N>::Vector(const T *d)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = d[i];
}

template<typename T, unsigned N>
inline Vector<T, N>::Vector(const T *d, unsigned stride)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = d[i*stride];
}

template<typename T, unsigned N>
template<typename... Args>
inline Vector<T, N>::Vector(T x_, Args... v)
{
	static_assert(1+sizeof...(v)==N, "Incorrect number of arguments in Vector constructor");
	(*this)[0] = x_;
	unsigned i = 1;
	for(auto c: std::initializer_list<T> { static_cast<T>(v)... })
		(*this)[i++] = c;
}

template<typename T, unsigned N>
template<typename U>
inline Vector<T, N>::Vector(const Vector<U, N> &v)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = v[i];
}

template<typename T, unsigned N>
inline Vector<T, N+1> compose(const Vector<T, N> &v, T s)
{
	Vector<T, N+1> r;
	for(unsigned i=0; i<N; ++i)
		r[i] = v[i];
	r[N] = s;
	return r;
}

template<typename T, unsigned N>
inline Vector<T, N+1> compose(T s, const Vector<T, N> &v)
{
	Vector<T, N+1> r;
	r[0] = s;
	for(unsigned i=0; i<N; ++i)
		r[i+1] = v[i];
	return r;
}

template<typename T, unsigned N, unsigned M>
inline Vector<T, N+M> compose(const Vector<T, N> &v1, const Vector<T, M> &v2)
{
	Vector<T, N+M> r;
	for(unsigned i=0; i<N; ++i)
		r[i] = v1[i];
	for(unsigned i=0; i<M; ++i)
		r[N+i] = v2[i];
	return r;
}

template<typename T, unsigned N>
template<unsigned M>
inline Vector<T, M> Vector<T, N>::slice(unsigned j) const
{
	Vector<T, M> r;
	for(unsigned i=0; i<M; ++i)
		r[i] = (*this)[j+i];
	return r;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator*=(T s)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] *= s;
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator*(const Vector<T, N> &v, T s)
{
	Vector<T, N> r(v);
	return r *= s;
}

template<typename T, unsigned N>
inline Vector<T, N> operator*(T s, const Vector<T, N> &v)
{
	return v*s;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator/=(T s)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] /= s;
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator/(const Vector<T, N> &v, T s)
{
	Vector<T, N> r(v);
	return r /= s;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator+=(const Vector<T, N> &v)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] += v[i];
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator+(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	Vector<T, N> r(v1);
	return r += v2;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator-=(const Vector<T, N> &v)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] -= v[i];
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator-(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	Vector<T, N> r(v1);
	return r -= v2;
}

template<typename T, unsigned N>
inline Vector<T, N> operator-(const Vector<T, N> &v)
{
	Vector<T, N> r(v);
	for(unsigned i=0; i<N; ++i)
		r[i] = -r[i];
	return r;
}

template<typename T, unsigned N>
inline bool operator==(const Vector<T, N> &v, const Vector<T, N> &w)
{
	for(unsigned i=0; i<N; ++i)
		if(v[i]!=w[i])
			return false;
	return true;
}

template<typename T, unsigned N>
inline T inner_product(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	T r = T();
	for(unsigned i=0; i<N; ++i)
		r += v1[i]*v2[i];
	return r;
}

template<typename T, unsigned N>
inline T Vector<T, N>::norm() const
{
	using std::sqrt;
	return sqrt(inner_product(*this, *this));
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::normalize()
{
	return *this /= norm();
}

template<typename T, unsigned N>
inline Vector<T, N> normalize(const Vector<T, N> &v)
{
	Vector<T, N> r(v);
	return r.normalize();
}

template<typename T>
inline T dot(const Vector<T, 3> &v1, const Vector<T, 3> &v2)
{
	return inner_product(v1, v2);
}

template<typename T>
inline Vector<T, 3> cross(const Vector<T, 3> &v1, const Vector<T, 3> &v2)
{
	return Vector<T, 3>(v1.y*v2.z-v1.z*v2.y, v1.z*v2.x-v1.x*v2.z, v1.x*v2.y-v1.y*v2.x);
}

template<typename T, unsigned N>
inline std::ostream &operator<<(std::ostream &s, const Vector<T, N> &v)
{
	s << "Vector" << N << '(';
	for(unsigned i=0; i<N; ++i)
	{
		if(i)
			s << ", ";
		s << v[i];
	}
	s << ')';
	return s;
}

} // namespace LinAl
} // namespace Msp

#endif
